/* atqgrammar.lex: a flex grammar for ATQ files
   Copyright (C)  2002-2008 Marco Cesati
   Thu, 02 Sep 2004 17:51:11 +0200
*/

%pointer
%option noyywrap stack
%x arg darg numarg tonumarg verbatim group filen filenskip

EOL     \r?\n
BLANKS  [\ \t]*
BLANKSNL  [\ \t]*{EOL}[\ \t]*
SLBEGIN \\begin{BLANKS}\{{BLANKS}
SLEND   \\end{BLANKS}\{{BLANKS}
ENDLINE {BLANKS}\}.*{EOL}
DIGIT   [0-9]
DIGITS  {DIGIT}+

%{
    #include "atqgrammar.h"
    int _yybrl = 0;
    int _value;
    int _linenum = 1;
    int _error;
    static int _obln;
    #define OPENBRACE do { _yybrl=1; _obln=_linenum; } while(0)
%} 

%%

    /********************* The recognized tokens *********************/
{BLANKS}\\mulquestion{BLANKS}\{       OPENBRACE;BEGIN(arg);    return MULQUESTION;
{BLANKS}\\question{BLANKS}\{          OPENBRACE;BEGIN(arg);    return QUESTION;
{BLANKS}\\groups{BLANKS}\{            OPENBRACE;BEGIN(group);  return GROUPS;
{BLANKS}\\course{BLANKS}\{            OPENBRACE;BEGIN(arg);    return COURSE;
{BLANKS}\\date{BLANKS}\{              OPENBRACE;BEGIN(arg);    return DATE;
{BLANKS}\\teacher{BLANKS}\{           OPENBRACE;BEGIN(arg);    return TEACHER;
{BLANKS}\\advise{BLANKS}\{            OPENBRACE;BEGIN(arg);    return ADVISE;
{BLANKS}\\good{BLANKS}\{              OPENBRACE;BEGIN(darg);   return GOOD;
{BLANKS}\\bad{BLANKS}\{               OPENBRACE;BEGIN(darg);   return BAD;
{BLANKS}\\fontpitch{BLANKS}\{         OPENBRACE;BEGIN(numarg); return FONTPITCH;
{BLANKS}\\atqfile{BLANKS}\{           OPENBRACE;BEGIN(filen);  return ATQFILE;
{BLANKS}\\preamble{BLANKS}\{          OPENBRACE;BEGIN(arg);    return PREAMBLE;
{BLANKS}\\formversion{BLANKS}\{       OPENBRACE;BEGIN(numarg); return FORMVERSION;
{BLANKS}\\formdate{BLANKS}\{          OPENBRACE;BEGIN(arg);    return FORMDATE;
{BLANKS}\\formcourse{BLANKS}\{        OPENBRACE;BEGIN(arg);    return FORMCOURSE;


    /* the splitted version ("\\token\n\{") */
{BLANKS}\\mulquestion{BLANKSNL}\{  OPENBRACE;++_linenum; BEGIN(arg);    return MULQUESTION;
{BLANKS}\\question{BLANKSNL}\{     OPENBRACE;++_linenum; BEGIN(arg);    return QUESTION;
{BLANKS}\\groups{BLANKSNL}\{       OPENBRACE;++_linenum; BEGIN(group);  return GROUPS;
{BLANKS}\\course{BLANKSNL}\{       OPENBRACE;++_linenum; BEGIN(arg);    return COURSE;
{BLANKS}\\date{BLANKSNL}\{         OPENBRACE;++_linenum; BEGIN(arg);    return DATE;
{BLANKS}\\teacher{BLANKSNL}\{      OPENBRACE;++_linenum; BEGIN(arg);    return TEACHER;
{BLANKS}\\advise{BLANKSNL}\{       OPENBRACE;++_linenum; BEGIN(arg);    return ADVISE;
{BLANKS}\\good{BLANKSNL}\{         OPENBRACE;++_linenum; BEGIN(darg);   return GOOD;
{BLANKS}\\bad{BLANKSNL}\{          OPENBRACE;++_linenum; BEGIN(darg);   return BAD;
{BLANKS}\\fontpitch{BLANKSNL}\{    OPENBRACE;++_linenum; BEGIN(numarg); return FONTPITCH;
{BLANKS}\\atqfile{BLANKSNL}\{      OPENBRACE;++_linenum; BEGIN(filen);  return ATQFILE;
{BLANKS}\\preamble{BLANKSNL}\{     OPENBRACE;++_linenum; BEGIN(arg);    return PREAMBLE;
{BLANKS}\\formversion{BLANKSNL}\{  OPENBRACE;++_linenum; BEGIN(numarg); return FORMVERSION;
{BLANKS}\\formdate{BLANKSNL}\{     OPENBRACE;++_linenum; BEGIN(arg);    return FORMDATE;
{BLANKS}\\formcourse{BLANKSNL}\{   OPENBRACE;++_linenum; BEGIN(arg);    return FORMCOURSE;

    /****** Verbatim environment is a sort of ``full-escape'' *******/
<verbatim>\\end\{verbatim\}           yy_pop_state(); yymore();
<verbatim>[^\\\n]*{EOL}               ++_linenum; yymore(); 
<verbatim>.                           yymore(); 


    /* LaTeX verbatim environment is peculiar */
<arg,darg>\\begin{BLANKS}\{verbatim\} yy_push_state(verbatim); yymore();

<arg,darg>[^{}\\\n]*{EOL}                yymore(); ++_linenum;
<arg,darg>[^{}\\\n]*                  yymore();
<arg,darg>\\.                         yymore();
<arg,darg>\{                          ++_yybrl; yymore();

<arg>\}                               if (!--_yybrl) {
                                          BEGIN(INITIAL);
                                          yytext[yyleng-1]='\0';
                                          return ARG;
                                      } else yymore();

<darg>\}                              if (!--_yybrl) {
                                          BEGIN(tonumarg);
                                          yytext[yyleng-1]='\0';
                                          return ARG;
                                      } else yymore();
                                      
    /******* Take care of second numeric argument of \bad and \good *******/ 
<tonumarg>{BLANKS}\{{BLANKS}          BEGIN(numarg);
<tonumarg>{BLANKS}{EOL}                  ++_linenum;
<tonumarg>.                           _error = 2; return ERR;


<numarg>{BLANKS}                      /* do nothing */
<numarg>{DIGITS}                      _value = atoi(yytext);
<numarg>{BLANKS}\}                    BEGIN(INITIAL); return NUM;
<numarg>{BLANKS}{EOL}                    ++_linenum;
<numarg>.                             _error = 0; return ERR;

    /******* The "groups" field has a peculiar format (a/b,c/d,e|f...) *****/
<group>{BLANKS}{DIGITS}               _value = atoi(yytext); return NUM;
<group>{BLANKS}[,/]                   /* do nothing */
<group>{BLANKS}\}                     BEGIN(INITIAL);
<group>.                              _error = 1; return ERR;

    /******* The file name argument of "atqfile" field **********************/

<filen>{BLANKS}[^{}\\\n]*              yymore();            
<filen>{BLANKS}\}                      BEGIN(filenskip); yytext[yyleng-1]='\0'; return ARG;
<filen>.                               _error = 4; return ERR;

<filenskip>.*{EOL}                     BEGIN(INITIAL); ++_linenum;

    /******* Comments start with a % and span till the end of line ********/
        /* <INITIAL,arg,darg,numarg,tonumarg,group,filen>\%.*{EOL}              ++_linenum; */
<INITIAL>\%.*{EOL}                     ++_linenum;

    /****** Blank lines between tokens are tolerated **********************/
{BLANKS}{EOL}                            ++_linenum;
{BLANKS}                                /* skip */

    /****** Any other char outside recognized tokens is an error **********/
[^\ \t\r\n]                             _error = 2; return ERR;

    /****** An EOF inside a token argument is definitely wrong ************/
<arg><<EOF>>                          |
<darg><<EOF>>                         |
<numarg><<EOF>>                       |
<verbatim><<EOF>>                     |
<filen><<EOF>>                        |
<group><<EOF>>                        _error = 3; _linenum=_obln; return ERR;
                                        
    /****** An EOF outside a token argument is fine ***********************/
<<EOF>>                               --_linenum; return 0;

%%

void restart_atq_scanner(FILE *input)
{
        _linenum = 1;
        _yybrl = 0;
        rewind(input);
        yyrestart(input);
        BEGIN(INITIAL);
}

