/* mkform.c -- Copyright (C)  2000-2008 Marco Cesati
 * Mon,  5 May 2003 18:38:28 +0200
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#if HAVE_UNISTD_H
#  include <unistd.h>
#endif
#if HAVE_STRING_H
#  include <string.h>
#endif
#if HAVE_SYS_TIME_H
#  include <time.h>
#endif
#if HAVE_SYS_TYPES_H
#  include <sys/types.h>
#endif

#if HAVE_STDLIB_H
#  include <stdlib.h>
#endif

#if HAVE_STRING_H
#  include <string.h>
#endif

#if HAVE_PWD_H
#  include <pwd.h>
#endif

#include "psform.h"

char const version[] = "mkform v0.7 [";

#define Abort() do { \
    fprintf(stderr, "Usage: %s [-f <num>] [-c \"coursename\"] " \
    "[-d \"exam date\"] [-a <size>[,<size]...] <num>\n", argv[0]); \
    exit(EXIT_FAILURE); \
} while (0)

static int get_num(char *arg)
{
    int l;
    char *p;
    
    l = (int) strtol(arg, &p, 10);
    if (*p)
        l = -1;
    return l; 
}

int get_ans_sizes(char *arg, int *len, int **arr)
{
    int i, l, v;
    char *p;
    
    if (!arg || !*arg)
        return 1;
        
    p = strchr(arg, ',');
    for (l=1; p; ++l)
        p = strchr(p+1, ',');

    *len = l;
    *arr = (int *) calloc(l, sizeof(int));
    
    if (!*arr) {
        fprintf(stderr, "Memory allocation error\n");
        exit(EXIT_FAILURE);
    }
    for (i=0, p=arg; i<l; ++i, ++p) {
        if (!p)
            return 1; /* paranoid (it should never occur) */
        v = (int) strtol(p, &p, 10);
        if (!v || (*p != ',' && *p != '\0'))
            return 1;
        (*arr)[i] = v;
    }
    return 0;
}

int main(int argc, char *argv[])
{
    int i, n, code, fnum;
    struct psform_desc_t fd;
    char date[26];
    const int USELEN = 64;
    char user[USELEN];
    const int CREATLEN = 512;
    char creator[CREATLEN];
    time_t t;
    uid_t u;
    struct passwd *p;
    
    if (argc == 1)
        Abort();

    memset(&fd, 0, sizeof(fd));
    
    fnum = 0;
    code = -1;

    fd.num_q = 0;
    fd.num_ans = NULL;
    
    strncpy(creator, version, CREATLEN);
    n = strlen(creator);
    for (i=0; i<argc; ++i) {
        if (n >= CREATLEN) {
            fprintf(stderr, "Parameters are too long\n");
            Abort();
        }
        n += snprintf(creator+n, CREATLEN-n, " %s", argv[i]);
        if (!i)
            continue;
        if (argv[i][0] == '-' && argv[i][1]) {
            char *parg, carg = argv[i][1];
            
            if (argv[i][2] == '\0') {
                n += snprintf(creator+n, CREATLEN-n, " %s", argv[i+1]);
                parg = argv[++i];
            } else
                parg = argv[i]+2;
            
            switch (carg) {
                case 'f':
                    fnum = get_num(parg);
                    break;
                case 'c':
                    fd.coursename = strdup(parg);
                    break;
                case 'd':
                    fd.examdate = strdup(parg);
                    break;
                case 'a':
                    if (!get_ans_sizes(parg, &fd.num_q, &fd.num_ans))
                        break;                        
                    fprintf(stderr, "Illegal argument to option \"-a\"\n");
                    /* FALLTHROUGH */
                default:
                    Abort();
            }
            continue;
        }
        fd.form_num = code = get_num(argv[i]);
        break;
    }

    strcat(creator, " ]");
    fd.creator = creator;
        
    if (i != argc-1)
        Abort();
    if (fnum < 0 || fnum > max_psform_version()) {
        fprintf(stderr, "Form version out of range (min 0, max %d)\n",
            max_psform_version());
        Abort();
    }
    if (code < 0 || code > 999) {
        fprintf(stderr, "Form number out of range (min 0, max 999)\n");
        Abort();
    }

    set_psform_version(fnum);
    
    if ((t = time(NULL)) == (time_t)-1) {
        fprintf(stderr, "%s: error reading time and date\n", argv[0]);
        t = 0;
    }

    if (!ctime_r(&t, date)) {
        fprintf(stderr, "%s: ctime_r() returned error\n", argv[0]);
        date[0] = '\0';
    } else
        date[strlen(date)-1] = '\0'; /* eat trailing newline */
    fd.date = date;

    u = getuid();
    p = getpwuid(u);
    if (p)
        sprintf(user, "%s@", p->pw_name);
    else
        sprintf(user, "[uid=%d]@", u);    

    if (gethostname(user+strlen(user),USELEN-strlen(user)-1)) {
        fprintf(stderr, "%s: gethostname() returned error\n", argv[0]);
        strcat(user, "[unknown host]");
    }
    fd.username = user;
     
    psform->print(&fd);

    return EXIT_SUCCESS;
}
