/* cross.c -- Copyright (C)  2000-2008 Marco Cesati
 * Fri, 18 Feb 2005 13:13:50 +0100
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#if HAVE_STDLIB_H
#  include <stdlib.h>
#endif
#if HAVE_STRING_H
#  include <string.h>
#endif

#include "form.h"
#include "rdform.h"

int find_max(int v[], int l)
{
    int i, m=-1, im=-1;
    
    for (i=0; i<l; ++i)
        if (v[i] > m)
            m=v[i], im=i;

    if (im != -1) {
        for (i=im+1; i<l; ++i)
            if (v[i] < m)
                break;
        im=(im+i-1)/2;
    }
    return im;    
}

static double find_cross_block(int top, int bottom, int left, int right,
                int *vertical, int *horizontal, int *size, struct image_t image)
{
    const double conf_hax = .40;
    const double conf_vax = .85;
    const double conf_box = .70;
    const double conf_prp = .90;
    int x, y;
    int h, v;
    int *vc, *hc;
    int tot, n, N, e;
    double confidence, hax, vax, box, prp;
    int i;
    int width = image.width;
    char *buf = image.buf;

#define PIXEL(L,O)  buf[(L)*width+(O)]

#if 0
    PRINT("[(%d,%d)-(%d,%d)] ", 
        left, top, right, bottom);
#endif
                                   
    hc = (int *) calloc(bottom-top+1, sizeof(int)); 
    vc = (int *) calloc(right-left+1, sizeof(int));

    if (!vc || !hc) {
        fprintf(stderr, "Memory allocation error.\n");
        exit(EXIT_FAILURE);
    }
             
    tot = 0;
    for (y = top; y<=bottom; ++y)
        for (x=left; x<=right; ++x)
            if (PIXEL(y,x)) {
                ++hc[y-top];
                ++vc[x-left];
                ++tot;
            }

    if (!tot)
        goto out;
    
    h = find_max(vc, right-left+1);
    v = find_max(hc,  bottom-top+1);
    
    if (h == -1 || v == -1)
        goto out;

    n = vc[h]+hc[v]-1; 

    if (vc[h] < 3 || hc[v] < 3)
        goto out;
        
    if (n < 20)
        goto out;
        
    x = left + h;
    y = top + v;

    e = (vc[h] > hc[v] ? vc[h] : hc[v]);
    
    for (i=y+1, N=0; i<=bottom; ++i)
        if (PIXEL(i,x))
            ++N;

    if (2*N>e)
        N=e-N-PIXEL(y,x);
    vax = (double)(2*N)/(e-1);
            
    for (i=x-1, N=0; i>=left; --i)
        if (PIXEL(y,i))
            ++N;

    if (2*N>e)
        N=e-N-PIXEL(y,x);

    hax = (double)(2*N)/(e-1);
    
    e = 2*e-1;

    N = (bottom-top+1)*(right-left+1);       
        
    if (N==n)
        prp = 0.0;
    else
        prp = (double)(N-tot)/(N-n);
        
    box = (double)n/e;   

    if (vax < conf_vax || hax < conf_hax || 
        prp < conf_prp || box < conf_box) {
out:
        free(hc);
        free(vc);
        return 0.0;
    }
    
    confidence = 100.0 *
                 (vax*hax*prp*box - conf_vax*conf_hax*conf_box*conf_prp) /
                 (1.0 - conf_vax*conf_hax*conf_box*conf_prp);

    *vertical = y;
    *horizontal = x;

#if 0
    PRINT("Cross mark: line %d, offset %d "
            "[%02.2f%%: %.3f %.3f %.3f %.3f]\n", 
            y, x, confidence, vax, hax, prp, box);
#endif


    free(vc);    
    free(hc);
    
    *size = n/2;
    return confidence;    
#undef PIXEL
}

/* Search the four crosses in the image; if successful, the function
 * fills the c_tlv, c_tlh, c_trv, c_trh, c_blv, c_blh, c_brv, c_brh image fields
 * with the x,y-coordinates of the four crosses 
 */

void find_cross_marks(struct image_t * pimage)
{  
#define HSTEPRATIO 3
#define VSTEPRATIO 3
    int width = pimage->width;
    int height = pimage->height;
    int hgap = width/16;
    int vgap = height/24;
    int x, y;
    double max;
    int xmax = 0, ymax = 0;
    int size;
     
    for (max=0.0, y=0; y<height/3 && max<70.0; y+=vgap/VSTEPRATIO)
        for (x=0; x<width/3; x+=hgap/HSTEPRATIO) {
            int tv, th;
            double t = find_cross_block(y, y+vgap, x, x+hgap, &tv, &th, &size, *pimage);
            if (t > max) {
                max = t;
                xmax = th;
                ymax = tv;
            }
        }        
    if (max == 0.0) {
        fprintf(stderr, "Top-left cross mark not found.\n");
        exit(EXIT_FAILURE);
    }
    
    pimage->c_tlv = ymax;
    pimage->c_tlh = xmax;
   
    PRINT("Top-left     cross mark found at (%4d,%4d), size=%3d, "
          "confidence=%6.2f%%\n", pimage->c_tlh, pimage->c_tlv, size, max);
    
    
    y = (pimage->c_tlv)-vgap/VSTEPRATIO;
    if (y < 0)
        y = 0;
        
    for (max=0.0, x=width-hgap-1; x>=(2*width)/3 && x>pimage->c_tlh; x-=hgap/HSTEPRATIO) {
        int tv, th;
        double t = find_cross_block(y, y+vgap, x, x+hgap, &tv, &th, &size, *pimage);
        if (t > max) {
            max = t;
            xmax = th;
            ymax = tv;
        }
    }
            
    if (max == 0.0) {
        for (y=0; y<height/3 && max<70.0; y+=vgap/VSTEPRATIO)
            for (x=width-hgap-1; x>=(2*width)/3 && x>pimage->c_tlh; x-=hgap/HSTEPRATIO) {
                int tv, th;
                double t = find_cross_block(y, y+vgap, x, x+hgap, &tv, 
                                            &th, &size, *pimage);
                if (t > max) {
                    max = t;
                    xmax = th;
                    ymax = tv;
                }
            }        
        if (max == 0.0) {
            fprintf(stderr, "Top-right cross mark not found.\n");
            exit(EXIT_FAILURE);
        }
    }        
    
    pimage->c_trv = ymax;
    pimage->c_trh = xmax;

    PRINT("Top-right    cross mark found at (%4d,%4d), size=%3d, "
            "confidence=%6.2f%%\n", pimage->c_trh, pimage->c_trv, size, max);
    
    for (max=0.0, y=height-vgap-1; y>=2*height/3 && max<70.0; y-=vgap/VSTEPRATIO)
        for (x=0; x<width/3; x+=hgap/HSTEPRATIO) {
            int tv, th;
            double t = find_cross_block(y, y+vgap, x, x+hgap, &tv, &th, &size, *pimage);
            if (t > max) {
                max = t;
                xmax = th;
                ymax = tv;
            }
        }

    if (max == 0.0) {
        fprintf(stderr, "Bottom-left cross mark not found.\n");
        exit(EXIT_FAILURE);
    }
    
    pimage->c_blv = ymax;
    pimage->c_blh = xmax;
    
    PRINT("Bottom-left  cross mark found at (%4d,%4d), size=%3d,"
        " confidence=%6.2f%%\n", pimage->c_blh, pimage->c_blv, size, max);

    y=(pimage->c_blv)-vgap/VSTEPRATIO;
    if (y+vgap >= height)
        y = height - vgap - 1;
    
    for (max=0.0, x=width-hgap-1; x>=(2*width)/3 && x>pimage->c_blh; x-=hgap/HSTEPRATIO) {
        int tv, th;
        double t = find_cross_block(y, y+vgap, x, x+hgap, &tv, &th, &size, *pimage);
        if (t > max) {
            max = t;
            xmax = th;
            ymax = tv;
        }
    }
       
    if (max == 0.0) {
        for (y=height-vgap-1; y>=2*height/3 && max<70.0; y-=vgap/VSTEPRATIO)
            for (x=width-hgap-1; x>=(2*width)/3 && x>pimage->c_blh; x-=hgap/HSTEPRATIO) {
                int tv, th;
                double t = find_cross_block(y, y+vgap, x, x+hgap, 
                                            &tv, &th, &size, *pimage);
                if (t > max) {
                    max = t;
                    xmax = th;
                    ymax = tv;
                }
            }
        if (max == 0.0) {
            fprintf(stderr, "Bottom-right cross mark not found.\n");
            exit(EXIT_FAILURE);
        }
    }
    
    pimage->c_brh = xmax;
    pimage->c_brv = ymax;
    
    PRINT("Bottom-right cross mark found at (%4d,%4d), size=%3d, "
            "confidence=%6.2f%%\n", pimage->c_brh, pimage->c_brv, size, max);

}
