/* ppm.c -- Copyright (C)  2000-2008 Marco Cesati
 * Fri, 18 Feb 2005 13:13:50 +0100
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#if HAVE_STDLIB_H
#  include <stdlib.h>
#endif

#include "rdform.h"

#define THRESHOLD   100000UL
#define WHITE       195075UL
#define BLACK            0UL

inline int read_line(FILE *f, unsigned char *line, int linesize)
{   
    int c;

    c = fread(line, 1, linesize, f);
    if (c == linesize)
        return 0;
    if (!c)
        return 1;
    if (feof(f))
        fprintf(stderr, "Truncated line in PPM file.\n");
    if (ferror(f))
        fprintf(stderr, "I/O error while reading PPM file.\n");
    return 1;
}

        

static void store_ppm_line(int nl, int *black, int *white, int *fuzzy,
            unsigned char *line, int linesize, char *buf, int width, int height)
{
    int i;
    int f = 0, w = 0, b = 0;

#define PIXEL(L,O)  buf[(L)*width+(O)]

    if (nl >= height) {
        fprintf(stderr, "Oops. Image includes too many lines.\n");
        exit(EXIT_FAILURE);
    }
    
    for (i=0; i<linesize; i+=3) {
        unsigned long  v = (line[i]*line[i])+
                           (line[i+1]*line[i+1])+
                           (line[i+2]*line[i+2]);

        if (v != WHITE && v != BLACK)
            f++;
        
        if (v > THRESHOLD) {
            w++;
            PIXEL(nl, i/3) = 0;          
        } else {
            b++;
            PIXEL(nl, i/3) = 1;
        }
    }
    *black += b;
    *white += w;
    *fuzzy += f;
#undef PIXEL
}

void read_ppm_file(const char *filename, struct image_t *pimage)
{
    FILE *f;
    char buffer[255];
    int nl;
    int white = 0, black = 0, fuzzy = 0;
    unsigned char *line;
    int linesize;
    int depth, width, height;
    char *image;
            
    f = fopen(filename, "r");
    if (!f) {
        perror("");
        exit(EXIT_FAILURE);
    }
    
    fgets(buffer, 255, f);
    if (buffer[0] != 'P' || (buffer[1] != '3' &&
                             buffer[1] != '6')) {
        fprintf(stderr, "Unknown format (not a PPM file)\n");
        exit(EXIT_FAILURE);    
    }    
    
    if (buffer[1] == '3') {
        fprintf(stderr, "ASCII PPM format not yet supported\n");
        exit(EXIT_FAILURE);
    }
    
    do {
        fgets(buffer, 255, f);
    } while (((buffer[0] == '#') || (buffer[0] == '\n')) && !feof(f));    

    if (sscanf(buffer, "%d %d", &width, &height) != 2) {
        fprintf(stderr, "Cannot get image size\n");
        exit(EXIT_FAILURE);
    }    

    fgets(buffer, 255, f);
    
    if (sscanf(buffer, "%d", &depth) != 1) {
        fprintf(stderr, "Cannot get image depth\n");
        exit(EXIT_FAILURE);
    }
    
    PRINT("PPM image %dx%d, max.col. %d\n", width, height, depth);
            
    if (depth!=255) {
        fprintf(stderr, "Pixel depth (max.col.=%d) is not supported.\n", depth);
        exit(EXIT_FAILURE);
    }

    linesize = width*3;
        
    line = (char *) malloc(linesize);
    
    if (!line) {
        fprintf(stderr, "Memory allocation error\n");
        exit(EXIT_FAILURE);
    }
    
    image = (char *) malloc(width*height);
    
    if (!image) {
        fprintf(stderr, "Memory allocation error\n");
        exit(EXIT_FAILURE);
    }

    nl = 0;
    while (!feof(f)) {
        if (read_line(f, line, linesize))
            break;
        store_ppm_line(nl, &black, &white, &fuzzy, line, linesize, 
                        image, width, height);
        nl++;
    }

    free(line);
    fclose(f);

    pimage->width = width;
    pimage->height = height;
    pimage->buf = image;

    PRINT("Pixels: black=%d white=%d fuzzy=%d\n", 
                    black, white, fuzzy);
    PRINT("Ink: %2.2f%% ", (float)black*100.0/(black+white));
    PRINT("Fuzziness: %2.2f%%\n", (float)fuzzy*100.0/(black+white));
} 
