/* rdform.c -- Copyright (C)  2000-2008 Marco Cesati
 * Sat, 19 Feb 2005 12:38:22 +0100
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#if HAVE_STDLIB_H
#  include <stdlib.h>
#endif
#include <math.h>
#if HAVE_STRING_H
#  include <string.h>
#endif

#include "form.h"
#include "rdform.h"

void dump_to_text(struct image_t image)
{
        int l,o;
        int width = image.width;
        int height = image.height;
        char *buf = image.buf;

#define PIXEL(L,O)  buf[(L)*width+(O)]

        PRINT("Output text file. Width=%d Height=%d\n", width, height);
        for (l=0;l<height;++l) {
            for (o=0; o<width; ++o)
                if (PIXEL(l,o))
                    putchar('#');
                else
                    putchar(' ');
            putchar('\n');
        }
#undef PIXEL
}

int round(double f)
{
    int n;
    double g;
    n = f;
    g = n;
    if (fabs(f-g) <= 0.5)
        return n;
    else
        return n+1.0;
}

void compute_resolution(struct image_t *pimage)
{

#define TRAPEZ_TOLERANCE_FACTOR   100

    double wdth, hght, phy;
    int h, v;
    double hres, vres;
    double c_htrapez, c_vtrapez;
    int c_tlh, c_blh, c_trh, c_brh, c_trv, c_tlv, c_brv, c_blv;
    
    find_cross_marks(pimage);
    
    c_tlh = pimage->c_tlh;
    c_tlv = pimage->c_tlv;
    c_blh = pimage->c_blh;
    c_blv = pimage->c_blv;
    c_trh = pimage->c_trh;
    c_trv = pimage->c_trv;
    c_brh = pimage->c_brh;
    c_brv = pimage->c_brv;
    
    if (c_tlh-c_blh-c_trh+c_brh || c_trv-c_tlv-c_brv+c_blv) {
        if (abs(c_tlh-c_blh-c_trh+c_brh) >
                                (c_trh-c_tlh)/TRAPEZ_TOLERANCE_FACTOR ||
            abs(c_trv-c_tlv-c_brv+c_blv) >
                                (c_brv-c_trv)/TRAPEZ_TOLERANCE_FACTOR) {
            fprintf(stderr, "\nCrosses do not appear properly placed.\n");
            exit(EXIT_FAILURE);
        }
    }    
    
    v = c_brv - c_blv + 1;
    h = c_brh - c_blh + 1;
    
    if (h <= 0) {
        fprintf(stderr, "\nEhm, your form is definitely too small!\n");
        exit(EXIT_FAILURE);
    }
        
    wdth = sqrt(h*h+v*v);
    
    PRINT("Form geometry: width=%2.2f pixels, ", wdth);
             
    h = c_blh - c_tlh + 1;
    v = c_blv - c_tlv + 1;
    
    if (v <= 0) {
        fprintf(stderr, "\nEhm, your form is definitely too small!\n");
        exit(EXIT_FAILURE);
    }
    
    hght = sqrt(h*h+v*v);
    
    PRINT("height=%2.2f pixels\n", hght);

    phy = atan((float)abs(c_blh-c_tlh)/(float)(c_blv-c_tlv));
    if (c_blh < c_tlh)
        phy = -phy;       
    
    pimage->cosphy=cos(phy);
    pimage->sinphy=sin(phy);

    PRINT("Form orientation: angle=%f degree (sin=%f, cos=%f)\n", 
        phy*180.0/M_PI, pimage->sinphy, pimage->cosphy);


    c_vtrapez = (double)(c_brv-c_trv-c_blv+c_tlv)/wdth;
    c_htrapez = (double)(c_trh-c_tlh-c_brh+c_blh)/hght;

    if (phy < 0.0) {
        c_vtrapez = -c_vtrapez;
        c_htrapez = -c_htrapez;
    }
     
    PRINT("Trapezoidal deformation: horizontal=%f vertical=%f\n", 
            c_htrapez, c_vtrapez);   
                                
    pimage->hres = hres = wdth/form->cross.hdist;
    pimage->vres = vres = hght/form->cross.vdist;
    
    PRINT("Resolution: horizontal=%.3f pixel/mm "
           "vertical=%.3f pixel/mm\n", hres, vres);
    
    if (fabs(hres-vres) > 1.0) {
        fprintf(stderr, "\nHorizontal and vertical resolutions differ.\n");
        exit(EXIT_FAILURE);
    }
    
}

static void form_offset(struct image_t image, double hmm, double vmm, int *hpix, int *vpix)
{
    double x, y, xp, yp;
    double hres = image.hres;
    double vres = image.vres;
    
    yp = vmm*vres;
    xp = hmm*hres;


    x = image.c_blh + (xp*image.cosphy - yp*image.sinphy);
    y = image.c_blv - (xp*image.sinphy + yp*image.cosphy);

#if 0
    x = image.c_blh + xp;
    y = image.c_blv - yp;
    if (image.c_tlv-image.c_blv)
        x += round(((float)(image.c_tlh-image.c_blh)*(float)(yp))/(float)(image.c_tlv-image.c_blv));
    if (image.c_brh-image.c_blh)
        y += round(((float)(image.c_brv-image.c_blv)*(float)(xp))/(float)(image.c_brh-image.c_blh));
    PRINT("(without phy) x=%f y=%f\n", x, y);   
#endif

#if 0
    x += round((width-x)*c_htrapez);
    y += round((height-y)*c_vtrapez);
#endif

    *hpix = round(x);
    *vpix = round(y);

}

static int code_mark(struct image_t image, int left, int top, double threshold)
{
    const double skip_edge = form->code.skip_edge;
    int x,y, tot, max;
    double hres = image.hres;
    double vres = image.vres;
    int hskip = round(skip_edge*hres);
    int vskip = round(skip_edge*vres);
    int bottom = top + round(form->code.height * vres) - vskip;
    int right = left + round(form->code.width * hres) - hskip;
    double ratio;
    int width = image.width;
    char *buf = image.buf;

#define PIXEL(L,O)  buf[(L)*width+(O)]

    top += vskip;
    left += hskip;
    max = (bottom-top+1)*(right-left+1);

     for (y=top, tot=0; y<=bottom; y++)
        for (x=left; x<=right; ++x)
            if (PIXEL(y,x))
                ++tot;

    ratio = (tot * 100.0) / max;

    PRINT("Code mark at (%4d,%4d): %3d pixels (%6.2f%%, %c)\n", 
            left, top, tot, ratio,
            (ratio >= threshold ? 'Y' : 'N'));

    return (ratio >= threshold);
#undef PIXEL
}

int extract_form_code(struct image_t image)
{
    const double max_threshold = form->code.max_threshold;
    const double min_threshold = form->code.min_threshold;
    const double step_threshold = form->code.step_threshold;
    int code[3];
    double *codeho = form->code.ho;
    double *codevo = form->code.vo;
    int i, j;
    double threshold;
    
    
    for (threshold = max_threshold; threshold >= min_threshold;
            threshold -= step_threshold) {
               
        code[0] = -1;
        code[1] = -1;
        code[2] = -1;
        
        PRINT("Searching code marks with threshold %6.2f%%\n", threshold);
        for (i=0; i<3; ++i) {
            for (j=0; j<10; ++j) {
                int x, y;
                form_offset(image, codeho[j], codevo[i], &x, &y);
                if (code_mark(image, x, y, threshold)) {
                    if (code[i] != -1) {
                        PRINT("Extra code[%d] (prev=%d now=%d)\n", 
                                i, code[i], j);
                        code[i] = -2;
                    } else
                        code[i] = j;
                }
            }
        }

        if (code[0] >= 0 && code[1] >= 0 && code[2] >= 0) {
            if (!quiet)
                fprintf(stderr, "Form code: %d%d%d (confidence %c= %2.0f%%)\n", 
                code[0], code[1], code[2], (threshold >= 100.0 ? '=' : '>'),
                threshold);
            return code[0]*100+code[1]*10+code[2];
        }

    }

    if (code[0] == -1 && code[1] == -1 && code[2] == -1) {
        PRINT("Null form code.\n");
        return -1;
    }

    PRINT("Invalid form code.\n");
    return -1;
}

static int test_mark(struct image_t image, int left, int top)
{    
    const int skip_edge = form->answer.skip_pixel;
    int x,y, tot, max;
    double hres = image.hres;
    double vres = image.vres;
    int bottom = top + round(form->answer.height * vres) - skip_edge;
    int right = left + round(form->answer.width * hres) - skip_edge;
    double ratio;
    int width = image.width;
    char *buf = image.buf;
    
#define PIXEL(L,O)  buf[(L)*width+(O)]

    top += skip_edge;
    left += skip_edge;
    max = (bottom-top+1)*(right-left+1);
    
    for (y=top, tot=0; y<=bottom; y++)
        for (x=left; x<=right; ++x)
            if (PIXEL(y,x))
                ++tot;
     
    ratio = (tot * 100.0) / max;
      
    PRINT("Box (%4d,%4d)-(%4d,%4d): %4d pixels (%6.2f%%, %c)\n", 
                left, top, right, bottom, tot, ratio, 
                (ratio >= form->answer.threshold ? 'Y' : 'N'));

    return (ratio >= form->answer.threshold);
#undef PIXEL
}

void extract_marks(char *vect, struct image_t image)
{
    int i, j;
    char marked;
    const double *testho = form->answer.ho;
    const double *testvo = form->answer.vo;
    int num_answer = form->answer.nanswer;
    int num_question = form->answer.nquest;

    PRINT("Searching boxes with threshold %6.2f%%\n", form->answer.threshold);    
    for (i=0; i<num_question; ++i)
        for (j=0, marked=0; j<num_answer; ++j) {
            int x, y;
            form_offset(image, testho[j], testvo[i], &x, &y);
            if (test_mark(image, x, y))
                marked |= (1 << j);
            vect[i] = marked;
        }
    return;
}
